@@ -18,6 +18,7 @@
 #include <sys/utsname.h>
 #include <sys/mman.h>
 
+#include "zlibsupport.h"
 #include "depmod.h"
 #include "moduleops.h"
 #include "tables.h"
@@ -27,6 +28,13 @@
 /* I hate strcmp. */
 #define streq(a,b) (strcmp((a),(b)) == 0)
 
+#ifndef MODULE_DIR
+#define MODULE_DIR "/lib/modules/"
+#endif
+
+static int verbose;
+static unsigned int skipchars;
+
 void fatal(const char *fmt, ...)
 {
 	va_list arglist;
@@ -166,7 +174,7 @@ static struct option options[] = { { "all", 0, NULL, 'a' },
 				   { "quick", 0, NULL, 'A' },
 				   { "basedir", 1, NULL, 'b' },
 				   { "errsyms", 0, NULL, 'e' },
-				   { "filesyms", 0, NULL, 'F' },
+				   { "filesyms", 1, NULL, 'F' },
 				   { "help", 0, NULL, 'h' },
 				   { "show", 0, NULL, 'n' },
 				   { "dry-run", 0, NULL, 'n' },
@@ -276,36 +284,27 @@ static struct module *grab_module(const char *dirname,
 				  const char *filename,
 				  struct module *next)
 {
-	struct stat buf;
 	struct module *new;
-	int fd;
 
 	new = NOFAIL(malloc(sizeof(*new)
 			    + strlen(dirname) + 1 + strlen(filename) + 1));
 	sprintf(new->pathname, "%s/%s", dirname, filename);
 	INIT_LIST_HEAD(&new->dep_list);
 
-	fd = open(new->pathname, O_RDONLY, 0);
-	if (fd < 0) {
+	new->data = grab_file(new->pathname, &new->len);
+	if (!new->data) {
 		warn("Can't read module %s: %s\n",
 		     new->pathname, strerror(errno));
-		goto fail;
-	}
-	fstat(fd, &buf);
-	new->mmap = mmap(0, buf.st_size, PROT_READ, MAP_PRIVATE, fd, 0);
-	if (new->mmap == MAP_FAILED) {
-		warn("Can't map module %s: %s\n",
-		     new->pathname, strerror(errno));
-		goto fail;
+		goto fail_data;
 	}
 
 	/* "\177ELF" <byte> where byte = 001 for 32-bit, 002 for 64 */
-	if (memcmp(new->mmap, ELFMAG, SELFMAG) != 0) {
+	if (memcmp(new->data, ELFMAG, SELFMAG) != 0) {
 		warn("Module %s is not an elf object\n", new->pathname);
 		goto fail;
 	}
 
-	switch (((char *)new->mmap)[EI_CLASS]) {
+	switch (((char *)new->data)[EI_CLASS]) {
 	case ELFCLASS32:
 		new->ops = &mod_ops32;
 		break;
@@ -314,19 +313,18 @@ static struct module *grab_module(const char *dirname,
 		break;
 	default:
 		warn("Module %s has elf unknown identifier %i\n",
-		     new->pathname, ((char *)new->mmap)[EI_CLASS]);
+		     new->pathname, ((char *)new->data)[EI_CLASS]);
 		goto fail;
 	}
 	new->ops->load_symbols(new);
 	new->ops->fetch_tables(new);
 
 	new->next = next;
-	close(fd);
 	return new;
 
- fail:
-	munmap(new->mmap, buf.st_size);
-	close(fd);
+fail:
+	release_file(new->data, new->len);
+fail_data:
 	free(new);
 	return next;
 }
@@ -417,15 +415,14 @@ static void del_module(struct module **modules, struct module *delme)
 {
 	struct module **i;
 
-	for (i = modules; *i; i = &(*i)->next)
-		if (*i == delme)
-			*i = delme->next;
+	/* Find pointer to it. */ 
+	for (i = modules; *i != delme; i = &(*i)->next);
+
+	*i = delme->next;
 }
 
 static void output_deps(struct module *modules,
-			unsigned int skipchars,
-			FILE *out,
-			int verbose)
+			FILE *out)
 {
 	struct module *i;
 
@@ -461,9 +458,7 @@ static void output_deps(struct module *modules,
 
 static int smells_like_module(const char *name)
 {
-	if (ends_in(name,".o") || ends_in(name,".klm") || ends_in(name,".ko"))
-		return 1;
-	return 0;
+	return ends_in(name,".ko") || ends_in(name, ".ko.gz");
 }
 
 static struct module *grab_dir(const char *dirname, struct module *next)
@@ -479,7 +474,6 @@ static struct module *grab_dir(const char *dirname, struct module *next)
 	}
 
 	while ((dirent = readdir(dir)) != NULL) {
-		/* Several competing extensions here.  .klm, .ko, .o */
 		if (smells_like_module(dirent->d_name))
 			next = grab_module(dirname, dirent->d_name, next);
 		else if (!streq(dirent->d_name, ".")
@@ -496,6 +490,7 @@ static struct module *grab_dir(const char *dirname, struct module *next)
 				next = grab_dir(subdir, next);
 		}
 	}
+	closedir(dir);
 	return next;
 }
 
@@ -522,7 +517,7 @@ static void filename2modname(char *modname, const char *filename)
 }
 
 /* Simply dump hash table. */
-static void output_symbols(FILE *out)
+static void output_symbols(struct module *unused, FILE *out)
 {
 	unsigned int i;
 
@@ -588,14 +583,74 @@ static void output_aliases(struct module *modules, FILE *out)
 	}
 }
 
+struct depfile {
+	char *name;
+	void (*func)(struct module *, FILE *);
+};
+
+static struct depfile depfiles[] = {
+	{ "modules.dep", output_deps }, /* This is what we check for '-A'. */
+	{ "modules.pcimap", output_pci_table },
+	{ "modules.usbmap", output_usb_table },
+	{ "modules.ccwmap", output_ccw_table },
+	{ "modules.ieee1394map", output_ieee1394_table },
+	{ "modules.alias", output_aliases },
+	{ "modules.symbols", output_symbols },
+};
+
+/* If we can't figure it out, it's safe to say "true". */
+static int any_modules_newer(const char *dirname, time_t mtime)
+{
+	DIR *dir;
+	struct dirent *dirent;
+
+	dir = opendir(dirname);
+	if (!dir)
+		return 1;
+
+	while ((dirent = readdir(dir)) != NULL) {
+		struct stat st;
+		char file[strlen(dirname) + 1 + strlen(dirent->d_name) + 1];
+
+		if (streq(dirent->d_name, ".") || streq(dirent->d_name, ".."))
+			continue;
+
+		sprintf(file, "%s/%s", dirname, dirent->d_name);
+		if (lstat(file, &st) != 0)
+			return 1;
+
+		if (smells_like_module(dirent->d_name)) {
+			if (st.st_mtime > mtime)
+				return 1;
+		} else if (S_ISDIR(st.st_mode)) {
+			if (any_modules_newer(file, mtime))
+				return 1;
+		}
+	}
+	closedir(dir);
+	return 0;
+}
+
+static int depfile_out_of_date(const char *dirname)
+{
+	struct stat st;
+	char depfile[strlen(dirname) + 1 + strlen(depfiles[0].name) + 1];
+
+	sprintf(depfile, "%s/%s", dirname, depfiles[0].name);
+
+	if (stat(depfile, &st) != 0)
+		return 1;
+
+	return any_modules_newer(dirname, st.st_mtime);
+}
+
 int main(int argc, char *argv[])
 {
-	int opt, all = 0, verbose = 0, bad_option = 0;
-	unsigned int skipchars = 0;
-	FILE *depout = NULL, *pciout = NULL, *usbout = NULL, *ieee1394out = NULL,
-		*ccwout = NULL, *aliasout = NULL, *symout = NULL;
-	char *basedir = "", *dirname, *version, *system_map = NULL;
+	int opt, all = 0, maybe_all = 0, doing_stdout = 0;
+	char *basedir = "", *dirname, *version, *badopt = NULL,
+		*system_map = NULL;
 	struct module *list = NULL;
+	int i;
 
 	/* Don't print out any errors just yet, we might want to exec
            backwards compat version. */
@@ -611,7 +666,7 @@ int main(int argc, char *argv[])
 			skipchars = strlen(basedir);
 			break;
 		case 'A':
-			all = 1;
+			maybe_all = 1;
 			break;
 		case 'F':
 			system_map = optarg;
@@ -631,14 +686,13 @@ int main(int argc, char *argv[])
 			print_usage(argv[0]);
 			break;
 		case 'n':
-			depout = pciout = usbout = ieee1394out = ccwout = aliasout = symout
-				= stdout;
+			doing_stdout = 1;
 			break;
 		case 'V':
 			printf("%s %s\n", PACKAGE, VERSION);
 			exit(0);
 		default:
-			bad_option = optopt;
+			badopt = argv[optind-1];
 		}
 	}
 
@@ -661,9 +715,9 @@ int main(int argc, char *argv[])
 	if (old_module_version(version))
 		exec_old_depmod(argv);
 
-	if (bad_option) {
-		fprintf(stderr, "%s: unrecognized option '%s'\n",
-			argv[0], argv[bad_option]);
+	if (badopt) {
+		fprintf(stderr, "%s: malformed/unrecognized option '%s'\n",
+			argv[0], badopt);
 		print_usage(argv[0]);
 		exit(1);
 	}
@@ -673,57 +727,14 @@ int main(int argc, char *argv[])
 		all = 1;
 
 	dirname = malloc(strlen(basedir)
-			 + strlen("/lib/modules/")
+			 + strlen(MODULE_DIR)
 			 + strlen(version) + 1);
-	sprintf(dirname, "%s%s%s", basedir, "/lib/modules/", version);
+	sprintf(dirname, "%s%s%s", basedir, MODULE_DIR, version);
 
-	/* If we're not stdout, open output files. */
-	if (!depout) {
-		/* Longest name to allocate */
-		char depname[strlen(dirname) + 1
-			    + sizeof("modules.generic_string")];
-
-		sprintf(depname, "%s/%s", dirname, "modules.dep");
-		depout = fopen(depname, "w");
-		if (!depout)
-			fatal("Could not open %s for writing: %s\n",
-			      depname, strerror(errno));
-
-		sprintf(depname, "%s/%s", dirname, "modules.pcimap");
-		pciout = fopen(depname, "w");
-		if (!pciout)
-			fatal("Could not open %s for writing: %s\n",
-			      depname, strerror(errno));
-
-		sprintf(depname, "%s/%s", dirname, "modules.usbmap");
-		usbout = fopen(depname, "w");
-		if (!usbout)
-			fatal("Could not open %s for writing: %s\n",
-			      depname, strerror(errno));
-
-		sprintf(depname, "%s/%s", dirname, "modules.ieee1394map");
-		ieee1394out = fopen(depname, "w");
-		if (!ieee1394out)
-			fatal("Could not open %s for writing: %s\n",
-			      depname, strerror(errno));
-
-		sprintf(depname, "%s/%s", dirname, "modules.ccwmap");
-		ccwout = fopen(depname, "w");
-		if (!ccwout)
-			fatal("Could not open %s for writing: %s\n",
-			      depname, strerror(errno));
-
-		sprintf(depname, "%s/%s", dirname, "modules.alias");
-		aliasout = fopen(depname, "w");
-		if (!aliasout)
-			fatal("Could not open %s for writing: %s\n",
-			      depname, strerror(errno));
-
-		sprintf(depname, "%s/%s", dirname, "modules.symbols");
-		symout = fopen(depname, "w");
-		if (!symout)
-			fatal("Could not open %s for writing: %s\n",
-			      depname, strerror(errno));
+	if (maybe_all) {
+		if (!doing_stdout && !depfile_out_of_date(dirname))
+			exit(0);
+		all = 1;
 	}
 
 	if (!all) {
@@ -734,13 +745,30 @@ int main(int argc, char *argv[])
 		list = grab_dir(dirname, list);
 	}
 
-	output_deps(list, skipchars, depout, verbose);
-	output_pci_table(list, pciout);
-	output_usb_table(list, usbout);
-	output_ccw_table(list, ccwout);
-	output_ieee1394_table(list, ieee1394out);
-	output_aliases(list, aliasout);
-	output_symbols(symout);
+	for (i = 0; i < sizeof(depfiles)/sizeof(depfiles[0]); i++) {
+		FILE *out;
+		struct depfile *d = &depfiles[i];
+		char depname[strlen(dirname) + 1 + strlen(d->name) + 1];
+		char tmpname[strlen(dirname) + 1 + strlen(d->name) +
+						strlen(".temp") + 1];
+
+		sprintf(depname, "%s/%s", dirname, d->name);
+		sprintf(tmpname, "%s/%s.temp", dirname, d->name);
+		if (!doing_stdout) {
+			out = fopen(tmpname, "w");
+			if (!out)
+				fatal("Could not open %s for writing: %s\n",
+					tmpname, strerror(errno));
+		} else
+			out = stdout;
+		d->func(list, out);
+		if (!doing_stdout) {
+			fclose(out);
+			if (rename(tmpname, depname) < 0)
+				fatal("Could not rename %s into %s: %s\n",
+					tmpname, depname, strerror(errno));
+		}
+	}
 
 	free(dirname);
 	free(version);
