@@ -41,8 +41,9 @@ struct module
 	unsigned int ccw_size;
 	void *ccw_table;
 
-	/* Memory map. */
-	void *mmap;
+	/* File contents and length. */
+	void *data;
+	unsigned long len;
 
 	char pathname[0];
 };
