@@ -39,55 +39,65 @@
 #include <sys/wait.h>
 #include <syslog.h>
 
+#include "zlibsupport.h"
 #include "list.h"
 #include "backwards_compat.c"
 
+extern long init_module(void *, unsigned long, const char *);
+extern long delete_module(const char *, unsigned int);
+
 struct module {
 	struct list_head list;
 	char *modname;
 	char filename[0];
 };
 
+#ifndef MODULE_DIR
 #define MODULE_DIR "/lib/modules"
+#endif
+
+typedef void (*errfn_t)(const char *fmt, ...);
 
 /* Do we use syslog or stderr for messages? */
 static int log;
+/* Should we be totally quiet? */
+static int quiet = 0;
 
-static int getlen(const char *fmt, va_list ap)
+static void message(const char *prefix, const char *fmt, va_list *arglist)
 {
-	int ret;
+	char *buf, *buf2;
 
-	/* glibcs before 2.1 return -1 on "can't fit". */
-	ret = vsnprintf(NULL, 0, fmt, ap);
-	if (ret >= 0)
-		return ret;
-	return 4096;
-}
+	if (quiet)
+		return;
 
-static void message(const char *prefix, const char *fmt, ...)
-{
-	char *buf;
-	va_list arglist;
-	int len;
-
-	va_start(arglist, fmt);
-	len = strlen(prefix) + getlen(fmt, arglist) + 1;
-	buf = malloc(len);
-
-	/* Must restart args on some archs */
-	va_start(arglist, fmt);
-	strcpy(buf, prefix);
-	vsnprintf(buf + strlen(buf), len, fmt, arglist);
+	vasprintf(&buf, fmt, *arglist);
+	asprintf(&buf2, "%s%s", prefix, buf);
 
 	if (log)
-		syslog(LOG_NOTICE, buf);
+		syslog(LOG_NOTICE, buf2);
 	else
-		fprintf(stderr, "%s", buf);
+		fprintf(stderr, "%s", buf2);
+	free(buf2);
+	free(buf);
+}
+
+static void warn(const char *fmt, ...)
+{
+	va_list arglist;
+	va_start(arglist, fmt);
+	message("WARNING: ", fmt, &arglist);
+	va_end(arglist);
+}
+
+static void fatal(const char *fmt, ...)
+{
+	va_list arglist;
+	va_start(arglist, fmt);
+	message("FATAL: ", fmt, &arglist);
+	va_end(arglist);
+	exit(1);
 }
 
-#define warn(fmt, ...) message("WARNING: ", fmt , ## __VA_ARGS__)
-#define fatal(fmt, ...) \
-	do { message("FATAL: ", fmt , ## __VA_ARGS__); exit(1); } while(0)
 
 static void grammar(const char *cmd, const char *filename, unsigned int line)
 {
@@ -353,23 +363,6 @@ static int module_in_kernel(const char *modname, int *usecount)
 	return 0;
 }
 
-static void *read_in(int fd, unsigned long len)
-{
-	void *ret = NOFAIL(malloc(len));
-	unsigned long done = 0;
-	int r;
-
-	while (done < len) {
-		r = read(fd, ret + done, len - done);
-		if (r <= 0) {
-			free(ret);
-			return NULL;
-		}
-		done += r;
-	}
-	return ret;
-}
-
 static void replace_modname(struct module *module,
 			    void *mem, unsigned long len,
 			    const char *oldname, const char *newname)
@@ -653,41 +646,45 @@ static char *add_extra_options(const char *modname,
 /* Do an install/remove command. */
 static void do_command(const char *modname,
 		       const char *command,
-		       int verbose, int dry_run, int dont_fail,
+		       int verbose, int dry_run,
+		       errfn_t error,
 		       const char *type)
 {
 	int ret;
 
 	if (verbose)
-		printf("%s %s: %s\n", type, modname, command);
+		printf("%s %s\n", type, command);
 	if (dry_run)
 		return;
 
 	ret = system(command);
-	if (ret == -1 || WEXITSTATUS(ret)) {
-		if (dont_fail)
-			fatal("Error running %s command for %s\n",
-			      type, modname);
-		warn("Error running %s command for %s\n", type, modname);
-	}
+	if (ret == -1 || WEXITSTATUS(ret))
+		error("Error running %s command for %s\n", type, modname);
 }
 
+enum exists_response {
+	EXISTS_ERROR,
+	EXISTS_SILENT_ERROR,
+	EXISTS_IGNORE,
+};
+
 /* Actually do the insert.  Frees second arg. */
 static void insmod(struct list_head *list,
 		   char *optstring,
 		   const char *newname,
-		   int dont_fail,
+		   enum exists_response exists,
+		   errfn_t error,
 		   int dry_run,
 		   int verbose,
 		   const struct module_options *options,
 		   const struct module_command *commands,
 		   int ignore_commands,
-		   int show_depends,
+		   int ignore_proc,
 		   int strip_vermagic,
 		   int strip_modversion)
 {
-	int fd, ret;
-	struct stat st;
+	int ret;
+	unsigned long len;
 	void *map;
 	const char *command;
 	struct module *mod = list_entry(list->next, struct module, list);
@@ -695,90 +692,82 @@ static void insmod(struct list_head *list,
 	/* Take us off the list. */
 	list_del(&mod->list);
 
-	/* Do things we depend on first. */
+	/* Do things we depend on first, but don't die if they fail. */
 	if (!list_empty(list)) {
-		insmod(list, NOFAIL(strdup("")), NULL, 0, dry_run, verbose,
-		       options, commands, show_depends, show_depends,
+		insmod(list, NOFAIL(strdup("")), NULL, EXISTS_IGNORE, warn,
+		       dry_run, verbose, options, commands, 0, ignore_proc,
 		       strip_vermagic, strip_modversion);
 	}
 
 	/* Did config file override command or add options? */
 	command = find_command(mod->modname, commands);
 	if (command && !ignore_commands) {
-		do_command(mod->modname, command, verbose, dry_run, dont_fail,
+		do_command(mod->modname, command, verbose, dry_run, error,
 			   "install");
 		goto out_optstring;
 	}
 
-	if (!show_depends
-	    && module_in_kernel(newname ?: mod->modname, NULL) == 1) {
-		if (dont_fail)
-			fatal("Module %s already in kernel.\n",
-			      newname ?: mod->modname);
-		goto out_optstring;
-	}
+	if (!ignore_proc
+	    && module_in_kernel(newname ?: mod->modname, NULL) == 1)
+		goto exists_error;
 
-	/* Open and suck in module (don't mmap, we might want to mangle) */
-	fd = open(mod->filename, O_RDONLY, 0);
-	if (fd < 0) {
-		if (dont_fail)
-			fatal("Could not open '%s': %s\n",
-			      mod->filename, strerror(errno));
-		warn("Could not open '%s': %s\n",
-		     mod->filename, strerror(errno));
-		goto out_optstring;
-	}
-
-	fstat(fd, &st);
-	map = read_in(fd, st.st_size);
+	map = grab_file(mod->filename, &len);
 	if (!map) {
-		if (dont_fail)
-			fatal("Can't read '%s': %s\n",
-			      mod->filename, strerror(errno));
-		warn("Can't read '%s': %s\n", mod->filename, strerror(errno));
-		goto out_fd;
+		error("Could not open '%s': %s\n",
+		      mod->filename, strerror(errno));
+		goto out_optstring;
 	}
 
 	/* Rename it? */
 	if (newname)
-		rename_module(mod, map, st.st_size, newname);
+		rename_module(mod, map, len, newname);
 
 	if (strip_modversion)
-		strip_section(mod, map, st.st_size, "__versions");
+		strip_section(mod, map, len, "__versions");
 	if (strip_vermagic)
-		clear_magic(mod, map, st.st_size);
+		clear_magic(mod, map, len);
 
 	/* Config file might have given more options */
 	optstring = add_extra_options(mod->modname, optstring, options);
 
-	if (show_depends)
-		printf("%s\n", mod->filename);
 	if (verbose)
 		printf("insmod %s %s\n", mod->filename, optstring);
 
 	if (dry_run)
 		goto out;
 
-	ret = syscall(__NR_init_module, map, st.st_size, optstring);
+	ret = init_module(map, len, optstring);
 	if (ret != 0) {
-		if (dont_fail)
-			fatal("Error inserting %s (%s): %s\n",
-			      mod->modname, mod->filename,
-			      insert_moderror(errno));
-		warn("Error inserting %s (%s): %s\n",
-		     mod->modname, mod->filename, insert_moderror(errno));
+		if (errno == EEXIST)
+			goto exists_error;
+		error("Error inserting %s (%s): %s\n",
+		      mod->modname, mod->filename, insert_moderror(errno));
 	}
  out:
-	free(map);
- out_fd:
-	close(fd);
+	release_file(map, len);
  out_optstring:
 	free(optstring);
+	return;
+
+exists_error:
+	if (exists == EXISTS_ERROR)
+		error("Module %s already in kernel.\n",
+		      newname ?: mod->modname);
+	else if (exists == EXISTS_SILENT_ERROR)
+		exit(1);
+	goto out_optstring;
 }
 
+enum noexists_response {
+	NOEXISTS_ERROR,
+	NOEXISTS_SILENT_ERROR,
+	NOEXISTS_IGNORE,
+};
+
 /* Do recursive removal. */
 static void rmmod(struct list_head *list,
-		  int dont_fail,
+		  enum noexists_response noexists,
+		  errfn_t error,
 		  int dry_run,
 		  int verbose,
 		  struct module_command *commands,
@@ -793,19 +782,17 @@ static void rmmod(struct list_head *list,
 
 	command = find_command(mod->modname, commands);
 	if (command && !ignore_commands) {
-		do_command(mod->modname, command, verbose, dry_run, dont_fail,
+		do_command(mod->modname, command, verbose, dry_run, error,
 			   "remove");
 		goto remove_rest;
 	}
 
-	if (module_in_kernel(mod->modname, &usecount) == 0) {
-		if (dont_fail)
-			fatal("Module %s is not in kernel.\n", mod->modname);
-		goto remove_rest;
-	}
+	if (module_in_kernel(mod->modname, &usecount) == 0)
+		goto nonexistent_module;
+
 	if (usecount != 0) {
-		if (dont_fail)
-			fatal("Module %s is in use.\n", mod->modname);
+		if (noexists != NOEXISTS_IGNORE)
+			error("Module %s is in use.\n", mod->modname);
 		goto remove_rest;
 	}
 
@@ -815,19 +802,27 @@ static void rmmod(struct list_head *list,
 	if (dry_run)
 		goto remove_rest;
 
-	if (syscall(__NR_delete_module, mod->modname, O_EXCL) != 0) {
-		if (dont_fail)
-			fatal("Error removing %s (%s): %s\n",
-			      mod->modname, mod->filename,
-			      remove_moderror(errno));
-		warn("Error removing %s (%s): %s\n",
-		     mod->modname, mod->filename, remove_moderror(errno));
+	if (delete_module(mod->modname, O_EXCL) != 0) {
+		if (errno == ENOENT)
+			goto nonexistent_module;
+		error("Error removing %s (%s): %s\n",
+		      mod->modname, mod->filename,
+		      remove_moderror(errno));
 	}
 
  remove_rest:
 	/* Now do things we depend. */
 	if (!list_empty(list))
-		rmmod(list, 0, dry_run, verbose, commands, 0);
+		rmmod(list, NOEXISTS_IGNORE, warn, dry_run, verbose, commands,
+		      0);
+	return;
+
+nonexistent_module:
+	if (noexists == NOEXISTS_ERROR)
+		fatal("Module %s is not in kernel.\n", mod->modname);
+	else if (noexists == NOEXISTS_SILENT_ERROR)
+		exit(1);
+	goto remove_rest;
 }
 
 static int do_modprobe(const char *argv0, const char *modname)
@@ -953,8 +948,7 @@ static char *strsep_skipspace(char **string, char *delim)
 	return strsep(string, delim);
 }
 
-/* FIXME: Maybe should be extended to "alias a b [and|or c]...".
-   modules.dep must also contain aliases extracted from modules. --RR */
+/* FIXME: Maybe should be extended to "alias a b [and|or c]...". --RR */
 
 /* Simple format, ignore lines starting with #, one command per line.
    Returns NULL or resolved alias. */
@@ -1010,7 +1004,9 @@ static char *read_config(const char *filename,
 				newresult = read_config(newfilename, 1, name,
 							dump_only, removing,
 							options, commands);
-				if (!result)
+				/* Files included override aliases,
+				   etc that was already set ... */
+				if (newresult)
 					result = newresult;
 			}
 		} else if (strcmp(cmd, "options") == 0) {
@@ -1090,6 +1086,26 @@ static char **merge_args(char *args, char *argv[], int *argc)
 	return newargs;
 }
 
+static char *gather_options(char *argv[])
+{
+	char *optstring = NOFAIL(strdup(""));
+
+	/* Rest is module options */
+	while (*argv) {
+		if (strchr(*argv, ' ')) {
+			/* Spaces handled by "" pairs, but no way of
+			   escaping quotes */
+			char protected_option[strlen(optstring) + 3];
+			sprintf(protected_option, "\"%s\"", *argv);
+			optstring = append_option(optstring, protected_option);
+		} else
+			optstring = append_option(optstring, *argv);
+		argv++;
+	}
+	return optstring;
+}
+
+
 static struct option options[] = { { "verbose", 0, NULL, 'v' },
 				   { "version", 0, NULL, 'V' },
 				   { "config", 1, NULL, 'C' },
@@ -1098,6 +1114,7 @@ static struct option options[] = { { "verbose", 0, NULL, 'v' },
 				   { "showconfig", 0, NULL, 'c' },
 				   { "autoclean", 0, NULL, 'k' },
 				   { "quiet", 0, NULL, 'q' },
+				   { "really-quiet", 0, NULL, 'Q' },
 				   { "show", 0, NULL, 'n' },
 				   { "dry-run", 0, NULL, 'n' },
 				   { "syslog", 0, NULL, 's' },
@@ -1114,6 +1131,7 @@ static struct option options[] = { { "verbose", 0, NULL, 'v' },
 				   { NULL, 0, NULL, 0 } };
 
 #define DEFAULT_CONFIG "/etc/modprobe.conf"
+#define MODPROBE_DEVFSD_CONF "/etc/modprobe.devfs"
 
 int main(int argc, char *argv[])
 {
@@ -1129,16 +1147,15 @@ int main(int argc, char *argv[])
 	int ignore_commands = 0;
 	int strip_vermagic = 0;
 	int strip_modversion = 0;
-	int fail_if_already = 1;
-	int show_depends = 0;
+	int ignore_proc = 0;
+	enum noexists_response noexists = NOEXISTS_ERROR;
+	enum exists_response exists = EXISTS_ERROR;
+	unsigned int i, num_modules;
 	char *type = NULL;
 	const char *config = NULL, *command = NULL;
-	char *dirname, *optstring = NOFAIL(strdup(""));
+	char *dirname, *optstring;
 	char *modname, *newname = NULL;
 	char *aliasfilename, *symfilename;
-	struct module_command *commands = NULL;
-	struct module_options *modoptions = NULL;
-	LIST_HEAD(list);
 
 	/* Prepend options from environment. */
 	argv = merge_args(getenv("MODPROBE_OPTIONS"), argv, &argc);
@@ -1153,7 +1170,7 @@ int main(int argc, char *argv[])
 		try_old_version("modprobe", argv);
 
 	uname(&buf);
-	while ((opt = getopt_long(argc, argv, "vVC:o:rknqsclt:aif", options, NULL)) != -1){
+	while ((opt = getopt_long(argc, argv, "vVC:o:rknqQsclt:aif", options, NULL)) != -1){
 		switch (opt) {
 		case 'v':
 			add_to_env_var("-v");
@@ -1173,8 +1190,8 @@ int main(int argc, char *argv[])
 			break;
 		case 'D':
 			dry_run = 1;
-			show_depends = 1;
-			ignore_commands = 1;
+			ignore_proc = 1;
+			verbose = 1;
 			add_to_env_var("-D");
 			break;
 		case 'o':
@@ -1202,9 +1219,18 @@ int main(int argc, char *argv[])
 			dry_run = 1;
 			break;
 		case 'q':
-			/* FIXME: should do more than this. */
+			/* install commands can rely on failure status
+			 * of children, so we only suppress output.
+			 *  install ip_conntrack /sbin/modprobe ip_conntrack
+			 *  && { /sbin/modprobe ip_conntrack_ftp; /bin/true; }
+			 */
 			add_to_env_var("-q");
-			fail_if_already = 0;
+			noexists = NOEXISTS_SILENT_ERROR;
+			exists = EXISTS_SILENT_ERROR;
+			break;
+		case 'Q':
+			add_to_env_var("-Q");
+			quiet = 1;
 			break;
 		case 's':
 			add_to_env_var("-s");
@@ -1234,11 +1260,17 @@ int main(int argc, char *argv[])
 		log = 1;
 	}
 
-	if (argc < optind + 1 && !dump_only && !list_only)
+	if (argc < optind + 1 && !dump_only && !list_only && !remove)
 		print_usage(argv[0]);
 
 	dirname = NOFAIL(malloc(strlen(buf.release) + sizeof(MODULE_DIR) + 1));
 	sprintf(dirname, "%s/%s", MODULE_DIR, buf.release);
+	aliasfilename = NOFAIL(malloc(strlen(dirname)
+				      + sizeof("/modules.alias")));
+	sprintf(aliasfilename, "%s/modules.alias", dirname);
+	symfilename = NOFAIL(malloc(strlen(dirname)
+				    + sizeof("/modules.symbols")));
+	sprintf(symfilename, "%s/modules.symbols", dirname);
 
 	/* Old-style -t xxx wildcard?  Or -l, or -a. */
 	if (all || type || list_only) {
@@ -1247,46 +1279,35 @@ int main(int argc, char *argv[])
 				   dry_run, verbose, argv[optind] ?: "*");
 	}
 
-	/* Rest is module options */
-	for (opt = optind + 1; opt < argc; opt++) {
-		if (strchr(argv[opt], ' ')) {
-			/* Spaces handled by "" pairs, but no way of
-			   escaping quotes */
-			char protected_option[strlen(optstring) + 3];
-			sprintf(protected_option, "\"%s\"", argv[opt]);
-			optstring = append_option(optstring, protected_option);
-		} else
-			optstring = append_option(optstring, argv[opt]);
+	if (remove) {
+		num_modules = argc - optind;
+		optstring = NOFAIL(strdup(""));
+		/* -r only allows certain restricted options */
+		if (newname)
+			fatal("Can't specify replacement name with -r\n");
+	} else {
+		num_modules = 1;
+		optstring = gather_options(argv+optind+1);
 	}
 
  	/* This is a horrible hack to allow devfsd, which calls
 	   modprobe with -C /etc/modules.conf or /etc/modules.devfs,
 	   to work.  FIXME. */
+	/* Modern devfsd or variants should use -Q explicitly in 2.6. */
 	if (config && !dump_only && strncmp(argv[optind], "/dev/", 5) == 0) {
-		if (strcmp("/etc/modules.conf", config) == 0)
+		if (strcmp("/etc/modules.conf", config) == 0) {
+			quiet = 1;
 			config = NULL;
-		else if (strcmp("/etc/modules.devfs", config) == 0)
-			config = "/etc/modprobe.devfs";
+		} else if (strcmp("/etc/modules.devfs", config) == 0) {
+			config = MODPROBE_DEVFSD_CONF;
+			quiet = 1;
+		}
 	}
 
-	/* -r only allows certain restricted options */
-	if (remove) {
-		if (strcmp(optstring, "") != 0)
-			fatal("Can't specify module options with -r\n");
-		if (newname)
-			fatal("Can't specify replacement name with -r\n");
-	}
-
-	/* Establish alias and symfilnames. */
-	aliasfilename = NOFAIL(malloc(strlen(dirname)
-				      + sizeof("/modules.alias")));
-	sprintf(aliasfilename, "%s/modules.alias", dirname);
-	symfilename = NOFAIL(malloc(strlen(dirname)
-				    + sizeof("/modules.symbols")));
-	sprintf(symfilename, "%s/modules.symbols", dirname);
-
-
 	if (dump_only) {
+		struct module_command *commands = NULL;
+		struct module_options *modoptions = NULL;
+
 		read_config(config ?: DEFAULT_CONFIG,
 			    config ? 1 : 0, "", 1, 0, &modoptions, &commands);
 		read_config(aliasfilename, 0, "", 1, 0,&modoptions, &commands);
@@ -1294,64 +1315,76 @@ int main(int argc, char *argv[])
 		exit(0);
 	}
 
-	/* Convert name we are looking for */
-	underscores(argv[optind]);
+	/* num_modules is always 1 except for remove */
+	for (i = 0; i < num_modules; i++) {
+		struct module_command *commands = NULL;
+		struct module_options *modoptions = NULL;
+		LIST_HEAD(list);
+		char *modulearg = argv[optind + i];
 
-	/* Returns the resolved alias, options */
-	modname = read_config(config ?: DEFAULT_CONFIG,
-			      config ? 1 : 0,
-			      argv[optind], 0,
-			      remove, &modoptions, &commands);
+		/* Convert name we are looking for */
+		underscores(modulearg);
 
-	/* No luck?  Try symbol names, if starts with symbol:. */
-	if (!modname && strncmp(argv[optind], "symbol:", strlen("symbol:")==0))
-		modname = read_config(symfilename, 0, argv[optind], 0,
+		/* Returns the resolved alias, options */
+		modname = read_config(config ?: DEFAULT_CONFIG,
+				      config ? 1 : 0,
+				      modulearg, 0,
 				      remove, &modoptions, &commands);
 
-	/* If we have an alias, gather any options associated with it
-           (needs to happen after parsing complete). */
-	if (modname) {
-	got_modname:
-		optstring = add_extra_options(argv[optind], optstring,
-					      modoptions);
-		read_depends(dirname, modname, &list);
-	} else {
-		read_depends(dirname, argv[optind], &list);
-		/* We don't allow canned aliases to override real
-		   modules, so we delay lookup until now. */
-		if (list_empty(&list)
-		    && (modname = read_config(aliasfilename, 0,
-					      argv[optind], 0,
-					      remove, &modoptions,
-					      &commands)))
-			goto got_modname;
+		/* No luck?  Try symbol names, if starts with symbol:. */
+		if (!modname
+		    && strncmp(modulearg, "symbol:", strlen("symbol:") == 0))
+			modname = read_config(symfilename, 0, modulearg, 0,
+					      remove, &modoptions, &commands);
 
-		modname = strdup(argv[optind]);
-	}
+		/* If we have an alias, gather any options associated with it
+		   (needs to happen after parsing complete). */
+		if (modname) {
+		got_modname:
+			optstring = add_extra_options(modulearg, optstring,
+						      modoptions);
+			read_depends(dirname, modname, &list);
+		} else {
+			read_depends(dirname, modulearg, &list);
+			/* We don't allow canned aliases to override real
+			   modules, so we delay lookup until now. */
+			if (list_empty(&list)
+			    && (modname = read_config(aliasfilename, 0,
+						      modulearg, 0,
+						      remove, &modoptions,
+						      &commands)))
+				goto got_modname;
 
-	if (list_empty(&list)) {
-		/* The dependencies have to be real modules, but
-		   handle case where the first is completely bogus. */
-		command = find_command(modname, commands);
-		if (command && !ignore_commands) {
-			do_command(modname, command, verbose, dry_run, 1,
-				   remove ? "remove" : "install");
-			goto out;
+			modname = strdup(modulearg);
 		}
-		fatal("Module %s not found.\n", modname);
+
+		if (list_empty(&list)) {
+			/* The dependencies have to be real modules, but
+			   handle case where the first is completely bogus. */
+			command = find_command(modname, commands);
+			if (command && !ignore_commands) {
+				do_command(modname, command, verbose, dry_run,
+					   fatal, remove ? "remove":"install");
+				continue;
+			}
+			if (noexists == NOEXISTS_ERROR)
+				fatal("Module %s not found.\n", modname);
+			else
+				exit(1);
+		}
+
+		if (remove)
+			rmmod(&list, noexists, fatal, dry_run, verbose,
+			      commands, ignore_commands);
+		else
+			insmod(&list, optstring, newname, exists, fatal,
+			       dry_run, verbose, modoptions, commands,
+			       ignore_commands, ignore_proc,
+			       strip_vermagic, strip_modversion);
+		free(modname);
 	}
 
-	if (remove)
-		rmmod(&list, fail_if_already, dry_run, verbose, commands,
-		      ignore_commands);
-	else
-		insmod(&list, optstring, newname, fail_if_already, dry_run,
-		       verbose, modoptions, commands, ignore_commands,
-		       show_depends, strip_vermagic, strip_modversion);
-
- out:
 	free(dirname);
-	free(modname);
 	free(aliasfilename);
 	free(symfilename);
 	if (log)
