@@ -29,6 +29,8 @@
 
 #include "backwards_compat.c"
 
+extern long delete_module(const char *, unsigned int);
+
 #define error(log, fmt, ...) message(log, "ERROR: ", fmt , ## __VA_ARGS__)
 #define warn(log, fmt, ...) message(log, "WARNING: ", fmt , ## __VA_ARGS__)
 #define info(log, fmt, ...) \
@@ -185,7 +187,7 @@ static void rmmod(int log, int verbose, const char *path, int flags)
 	     (flags & O_NONBLOCK) ? "no" : "yes",
 	     (flags & O_TRUNC) ? " force" : "");
 
-	ret = syscall(__NR_delete_module, name, flags);
+	ret = delete_module(name, flags);
 	if (ret != 0)
 		error(log, "Removing '%s': %s\n", name, strerror(errno));
 }
